=== modified file 'bzrlib/graph_walker.py'
--- bzrlib/graph_walker.py	2007-05-25 04:44:29 +0000
+++ bzrlib/graph_walker.py	2007-05-26 00:20:09 +0000
@@ -24,13 +24,26 @@
             self._ancestors[ghost] = [NULL_REVISION]
 
     def distance_from_origin(self, revisions):
+        """Determine the of the named revisions from the origin
+
+        :param revisions: The revisions to examine
+        :return: A list of revision distances.  None is provided if no distance
+            could be found.
+        """
         distances = graph.node_distances(self._descendants, self._ancestors,
                                          NULL_REVISION)
         return [distances.get(r) for r in revisions]
 
-    def minimal_common(self, left_revision, right_revision):
-        common = set(self._get_ancestry(left_revision))
-        common.intersection_update(self._get_ancestry(right_revision))
+    def minimal_common(self, *revisions):
+        """Determine the minimal common ancestors of the provided revisions
+
+        A minimal common ancestor is a common ancestor none of whose
+        descendants are common ancestors.  (This is not quite the standard
+        graph theory definition)
+        """
+        common = set(self._get_ancestry(revisions[0]))
+        for revision in revisions[1:]:
+            common.intersection_update(self._get_ancestry(revision))
         common.add(NULL_REVISION)
         mca = set()
         for ancestor in common:
@@ -39,6 +52,22 @@
                 mca.add(ancestor)
         return mca
 
+    def unique_common(self, left_revision, right_revision):
+        """Find a unique minimal common ancestor.
+
+        Find minimal common ancestors.  If there is no unique minimal common
+        ancestor, find the minimal common ancestors of those ancestors.
+
+        Iteration stops when a unique minimal common ancestor is found.
+        The graph origin is necessarily a unique common ancestor
+        """
+        revisions = [left_revision, right_revision]
+        while True:
+            minimal = self.minimal_common(*revisions)
+            if len(minimal) == 1:
+                return minimal.pop()
+            revisions = minimal
+
     def _get_ancestry(self, revision):
         if revision == NULL_REVISION:
             ancestry = []

=== modified file 'bzrlib/tests/test_graph_walker.py'
--- bzrlib/tests/test_graph_walker.py	2007-05-25 04:44:29 +0000
+++ bzrlib/tests/test_graph_walker.py	2007-05-26 00:20:09 +0000
@@ -4,15 +4,21 @@
 ancestry_1 = {'rev1': [NULL_REVISION], 'rev2a': ['rev1'], 'rev2b': ['rev1'],
               'rev3': ['rev2a'], 'rev4': ['rev3', 'rev2b']}
 
+criss_cross = {'rev1': [NULL_REVISION], 'rev2a': ['rev1'], 'rev2b': ['rev1'],
+               'rev3a': ['rev2a', 'rev2b'], 'rev3b': ['rev2b', 'rev2a']}
+
 class TestGraphWalker(TestCaseWithMemoryTransport):
 
     def test_distance_from_origin(self):
-        tree = self.build_ancestry(ancestry_1)
-        graph_walker = tree.branch.repository.get_graph_walker()
+        graph_walker = self.make_walker(ancestry_1)
         self.assertEqual([1, 0, 2, 4],
                          graph_walker.distance_from_origin(['rev1', 'null:',
                          'rev2b', 'rev4']))
 
+    def make_walker(self, ancestors):
+        tree = self.build_ancestry(ancestors)
+        return tree.branch.repository.get_graph_walker()
+
     def build_ancestry(self, ancestors):
         tree = self.make_branch_and_memory_tree('.')
         tree.lock_write()
@@ -25,26 +31,76 @@
         while len(pending) > 0:
             cur_node = pending.pop()
             for descendant in descendants.get(cur_node, []):
+                if tree.branch.repository.has_revision(descendant):
+                    continue
                 parents = [p for p in ancestors[descendant] if p is not
                            NULL_REVISION]
                 if len([p for p in parents if not
                     tree.branch.repository.has_revision(p)]) > 0:
                     continue
                 tree.set_parent_ids(parents)
+                if len(parents) > 0:
+                    left_parent = parents[0]
+                else:
+                    left_parent = NULL_REVISION
                 tree.branch.set_last_revision_info(
-                    len(tree.branch._lefthand_history(cur_node)),
-                    cur_node)
+                    len(tree.branch._lefthand_history(left_parent)),
+                    left_parent)
                 tree.commit(descendant, rev_id=descendant)
                 pending.append(descendant)
         tree.unlock()
         return tree
 
     def test_mca(self):
-        tree = self.build_ancestry(ancestry_1)
-        graph_walker = tree.branch.repository.get_graph_walker()
-        self.assertEqual(set([NULL_REVISION]),
-                         graph_walker.minimal_common(NULL_REVISION,
-                                                     NULL_REVISION))
-        self.assertEqual(set([NULL_REVISION]),
-                         graph_walker.minimal_common(NULL_REVISION,
-                                                     'rev1'))
+        """Test finding minimal common ancestor.
+
+        ancestry_1 should always have a single common ancestor
+        """
+        graph_walker = self.make_walker(ancestry_1)
+        self.assertEqual(set([NULL_REVISION]),
+                         graph_walker.minimal_common(NULL_REVISION,
+                                                     NULL_REVISION))
+        self.assertEqual(set([NULL_REVISION]),
+                         graph_walker.minimal_common(NULL_REVISION,
+                                                     'rev1'))
+        self.assertEqual(set(['rev1']),
+                         graph_walker.minimal_common('rev1', 'rev1'))
+        self.assertEqual(set(['rev1']),
+                         graph_walker.minimal_common('rev2a', 'rev2b'))
+
+    def test_mca_criss_cross(self):
+        graph_walker = self.make_walker(criss_cross)
+        self.assertEqual(set(['rev2a', 'rev2b']),
+                         graph_walker.minimal_common('rev3a', 'rev3b'))
+        self.assertEqual(set(['rev2b']),
+                         graph_walker.minimal_common('rev3a', 'rev3b',
+                                                     'rev2b'))
+
+    def test_recursive_unique_mca(self):
+        """Test finding a unique minimal common ancestor.
+
+        ancestry_1 should always have a single common ancestor
+        """
+        graph_walker = self.make_walker(ancestry_1)
+        self.assertEqual(NULL_REVISION,
+                         graph_walker.unique_common(NULL_REVISION,
+                                                     NULL_REVISION))
+        self.assertEqual(NULL_REVISION,
+                         graph_walker.unique_common(NULL_REVISION,
+                                                     'rev1'))
+        self.assertEqual('rev1', graph_walker.unique_common('rev1', 'rev1'))
+        self.assertEqual('rev1', graph_walker.unique_common('rev2a', 'rev2b'))
+
+    def test_mca_criss_cross(self):
+        graph_walker = self.make_walker(criss_cross)
+        self.assertEqual(set(['rev2a', 'rev2b']),
+                         graph_walker.minimal_common('rev3a', 'rev3b'))
+        self.assertEqual(set(['rev2b']),
+                         graph_walker.minimal_common('rev3a', 'rev3b',
+                                                     'rev2b'))
+
+    def test_unique_common_criss_cross(self):
+        """Ensure we don't pick non-unique mcas in a criss-cross"""
+        graph_walker = self.make_walker(criss_cross)
+        self.assertEqual('rev1',
+                         graph_walker.unique_common('rev3a', 'rev3b'))

